"""
Contains GWS DAO implementations.
"""
from resttools.mock.mock_http import MockHTTP
from resttools.dao_implementation.live import get_con_pool, get_live_url
from resttools.dao_implementation.mock import get_mockdata_url


class File(object):
    """
    The File DAO implementation returns generally static content.  Use this
    DAO with this configuration:

    RESTTOOLS_GWS_DAO_CLASS = 'resttools.dao_implementation.gws.File'
    """
    _max_pool_size = 5
    _cache_db = {}

    def __init__(self, conf):
        self._conf = conf
        if 'MAX_POOL_SIZE' in conf:
            self._max_pool_size = conf['MAX_POOL_SIZE']

    def getURL(self, url, headers):
        if url not in File._cache_db:
            response = get_mockdata_url("gws", self._conf, url, headers)
        else:
            response = MockHTTP()
            response.data = File._cache_db[url]
            response.status = 200
        return response

    def putURL(self, url, headers, body):
        response = MockHTTP()

        if body is not None:
            if "If-Match" in headers:
                response.status = 200  # update
            else:
                response.status = 201  # create
            response.headers = {"X-Data-Source": "GWS file mock data",
                                "Content-Type": headers["Content-Type"]}
            response.data = body
        else:
            response.status = 400
            response.data = "Bad Request: no POST body"

        return response

    def deleteURL(self, url, headers):
        response = MockHTTP()
        response.status = 200
        return response


class Live(object):
    """
    This DAO provides real data.  It requires further configuration, (conf)
    """
    _max_pool_size = 5
    _socket_timeout = 20.0

    def __init__(self, conf):
        self._conf = conf
        if 'MAX_POOL_SIZE' in conf:
            self._max_pool_size = conf['MAX_POOL_SIZE']
        if 'SOCKET_TIMEOUT' in conf:
            self._socket_timeout = conf['SOCKET_TIMEOUT']

    pool = None

    def getURL(self, url, headers):
        if Live.pool is None:
            Live.pool = self._get_pool()

        return get_live_url(Live.pool, 'GET',
                            self._conf['HOST'],
                            url, headers=headers,
                            service_name='gws')

    def putURL(self, url, headers, body):
        if Live.pool is None:
            Live.pool = self._get_pool()

        return get_live_url(Live.pool, 'PUT',
                            self._conf['HOST'],
                            url, headers=headers, body=body,
                            service_name='gws')

    def deleteURL(self, url, headers):
        if Live.pool is None:
            Live.pool = self._get_pool()

        return get_live_url(Live.pool, 'DELETE',
                            self._conf['HOST'],
                            url, headers=headers,
                            service_name='gws')

    def _get_pool(self):
        return get_con_pool(self._conf['HOST'],
                            self._conf['KEY_FILE'],
                            self._conf['CERT_FILE'],
                            self._conf['CA_FILE'],
                            socket_timeout=self._socket_timeout,
                            max_pool_size=self._max_pool_size)



"""
Contains NWS DAO implementations.
"""

from resttools.mock.mock_http import MockHTTP
import re
from resttools.dao_implementation.live import get_con_pool, get_live_url
from resttools.dao_implementation.mock import get_mockdata_url

import logging
logger = logging.getLogger(__name__)


class File(object):
    """
    The File DAO implementation returns generally static content.  Use this
    DAO with this configuration:

    """
    _max_pool_size = 5
    _cache_db = {}

    def __init__(self, conf):
        self._conf = conf
        if 'MAX_POOL_SIZE' in conf:
            self._max_pool_size = conf['MAX_POOL_SIZE']

    def getURL(self, url, headers):
        logger.debug('file nws get url: ' + url)
        if url not in File._cache_db:
            response = get_mockdata_url("nws", self._conf, url, headers)
        else:
            logger.debug('using cache')
            response = MockHTTP()
            response.data = File._cache_db[url]
            response.status = 200
        if response.status == 404:
            logger.debug('status 404')
            response.data = '{"error": {"code": "7000","message": "No record matched"}}'
        return response

    def postURL(self, url, headers, body):
        logger.debug('file nws post url: ' + url)

        response = get_mockdata_url("nws", self._conf, url, headers)
        if response.status == 404:
            logger.debug('status 404')
            response.data = '{"error": {"code": "7000","message": "No record matched"}}'

        return response


class Live(object):
    """
    This DAO provides real data.  It requires further configuration, (conf)
    """
    _max_pool_size = 5

    def __init__(self, conf):
        self._conf = conf
        if 'MAX_POOL_SIZE' in conf:
            self._max_pool_size = conf['MAX_POOL_SIZE']

    pool = None

    def getURL(self, url, headers):
        if Live.pool is None:
            Live.pool = self._get_pool()

        return get_live_url(Live.pool, 'GET',
                            self._conf['HOST'],
                            url, headers=headers,
                            service_name='nws')

    def postURL(self, url, headers, body):
        if Live.pool is None:
            Live.pool = self._get_pool()

        return get_live_url(Live.pool, 'POST',
                            self._conf['HOST'],
                            url, headers=headers, body=body,
                            service_name='nws')

    def _get_pool(self):
        return get_con_pool(self._conf['HOST'],
                            self._conf['KEY_FILE'],
                            self._conf['CERT_FILE'],
                            self._conf['CA_FILE'],
                            max_pool_size=self._max_pool_size, verify_https=False)



# resttools implementation for non-django applications
from resttools.dao_implementation.irws import File as IRWSFile
from resttools.dao_implementation.irws import Live as IRWSLive
from resttools.dao_implementation.nws import File as NWSFile
from resttools.dao_implementation.nws import Live as NWSLive
from resttools.dao_implementation.gws import File as GWSFile
from resttools.dao_implementation.gws import Live as GWSLive
from resttools.dao_implementation.ntfyws import File as NTFYWSFile
from resttools.dao_implementation.ntfyws import Live as NTFYWSLive


class DAO_BASE(object):

    def __init__(self, conf):
        self._conf = conf
        self._run_mode = conf['RUN_MODE']

    def _getURL(self, service, url, headers):
        dao = self._getDAO()
        response = dao.getURL(url, headers)
        return response

    def _postURL(self, service, url, headers, body=None):
        dao = self._getDAO()
        response = dao.postURL(url, headers, body)
        return response

    def _deleteURL(self, service, url, headers):
        dao = self._getDAO()
        response = dao.deleteURL(url, headers)
        return response

    def _putURL(self, service, url, headers, body=None):
        dao = self._getDAO()
        response = dao.putURL(url, headers, body)
        return response


class IRWS_DAO(DAO_BASE):
    def getURL(self, url, headers):
        return self._getURL('irws', url, headers)

    def putURL(self, url, headers, body):
        return self._putURL('irws', url, headers, body)

    def postURL(self, url, headers, body):
        return self._postURL('irws', url, headers, body)

    def deleteURL(self, url, headers):
        return self._deleteURL('irws', url, headers)

    def _getDAO(self):
        if self._run_mode == 'Live':
            return IRWSLive(self._conf)
        return IRWSFile(self._conf)


class NWS_DAO(DAO_BASE):
    def getURL(self, url, headers):
        return self._getURL('nws', url, headers)

    def postURL(self, url, headers, body):
        return self._postURL('nws', url, headers, body)

    def _getDAO(self):
        if self._run_mode == 'Live':
            return NWSLive(self._conf)
        return NWSFile(self._conf)


class GWS_DAO(DAO_BASE):
    def getURL(self, url, headers):
        return self._getURL('gws', url, headers)

    def putURL(self, url, headers, body):
        return self._putURL('gws', url, headers, body)

    def deleteURL(self, url, headers):
        return self._deleteURL('gws', url, headers)

    def _getDAO(self):
        if self._run_mode == 'Live':
            return GWSLive(self._conf)
        return GWSFile(self._conf)


class NTFYWS_DAO(DAO_BASE):
    def postURL(self, url, headers, body):
        return self._postURL('ntfyws', url, headers, body)

    def _getDAO(self):
        if self._run_mode == 'Live':
            return NTFYWSLive(self._conf)
        return NTFYWSFile(self._conf)



# config for tests

import os

DEBUG = False

RUN_MODE = 'File'
MOCK_ROOT = os.path.abspath(os.path.dirname(__file__)) + '/test_data'

#

GWS_CONF = {
    'HOST':  'https://iam-ws.u.washington.edu:7443',
    'MAX_POOL_SIZE': 5,
}

IRWS_CONF = {
    'HOST':  'https://mango-dev.u.washington.edu:646',
    'SERVICE_NAME': 'registry-dev',
    'MAX_POOL_SIZE': 5,
    'VERIFY_HOST': False,
}

NWS_CONF = {
    'HOST':  'https://uwnetid.washington.edu',
    'SERVICE_NAME': 'nws',
    'MAX_POOL_SIZE': 5,
    'PASSWORD_ACTION': 'Test',
}

NTFYWS_CONF = {
    'HOST':  'https://notify-dev.s.uw.edu',
    # 'HOST':  'https://notify.uw.edu',
    'SERVICE_NAME': 'notification',
    'MOCK_ROOT': 'nsspr/mock_data',
}


# augment handlers' configs
for conf in (GWS_CONF, IRWS_CONF, NWS_CONF):
    if 'RUN_MODE' not in conf:
        conf['RUN_MODE'] = RUN_MODE
    if 'MOCK_ROOT' not in conf:
        conf['MOCK_ROOT'] = MOCK_ROOT

LOGGING = {
    'version': 1,
    'formatters': {
        'plain': {
            'format': '%(message)s'
        },
    },
    'handlers': {
        'console': {
            'class': 'logging.StreamHandler',
            'level': 'DEBUG',
            'stream': 'ext://sys.stdout',
            'formatter': 'plain',
        },
        'default': {
            'level': 'DEBUG',
            'class': 'logging.FileHandler',
            'formatter': 'plain',
            'filename': 'tests.log',
        },
    },
    'root': {
        'level': 'DEBUG',
        'handlers': ['default'],
    },
    'suds': {
        'level': 'INFO',
        'handlers': ['default']
    },
}



"""
Provides access to the http connection pools and
connections for live data from a web service

"""
import ssl
from six.moves.urllib.parse import urlparse
from urllib3 import connection_from_url


def get_con_pool(host,
                 key_file=None,
                 cert_file=None,
                 ca_file=None,
                 socket_timeout=15.0,
                 max_pool_size=3,
                 verify_https=True):
    """
    Return a ConnectionPool instance of given host
    :param socket_timeout:
        socket timeout for each connection in seconds
    """
    kwargs = {
        "timeout": socket_timeout,
        "maxsize": max_pool_size,
        "block": True,
        }

    if key_file is not None and cert_file is not None:
        kwargs["key_file"] = key_file
        kwargs["cert_file"] = cert_file

    if urlparse(host).scheme == "https":
        kwargs["ssl_version"] = ssl.PROTOCOL_TLSv1
        if verify_https:
            kwargs["cert_reqs"] = "CERT_REQUIRED"
            kwargs["ca_certs"] = ca_file

    return connection_from_url(host, **kwargs)


def get_live_url(con_pool,
                 method,
                 host,
                 url,
                 headers,
                 retries=3,
                 body=None,
                 service_name=None):
    """
    Return a connection from the pool and perform an HTTP request.
    :param con_pool:
        is the http connection pool associated with the service
    :param method:
        HTTP request method (such as GET, POST, PUT, etc.)
    :param host:
        the url of the server host.
    :param headers:
        headers to include with the request
    :param body:
        the POST, PUT body of the request
    """
    timeout = con_pool.timeout
    response = con_pool.urlopen(method, url, body=body, headers=headers, retries=retries, timeout=timeout)
    return response



"""
This is the interface for interacting with the Group Web Service.
"""
from resttools.dao import GWS_DAO
from resttools.exceptions import InvalidGroupID
from resttools.exceptions import DataFailureException
from resttools.models.gws import Group, CourseGroup, GroupReference
from resttools.models.gws import GroupUser, GroupMember
from six.moves.urllib.parse import urlencode
from lxml import etree
import re
from jinja2 import Environment, PackageLoader

import logging
logger = logging.getLogger(__name__)


class GWS(object):
    """
    The GWS object has methods for getting group information.
    """
    def __init__(self, conf, actas=None):
        self._service_name = 'gws'
        self._conf = conf
        self._j2env = Environment(loader=PackageLoader('resttools',
                                                       'templates/gws'))
        self._actas = actas
        self.dao = GWS_DAO(conf)

    QTRS = {'win': 'winter', 'spr': 'spring', 'sum': 'summer', 'aut': 'autumn'}

    def search_groups(self, **kwargs):
        """
        Returns a list of resttools.GroupReference objects matching the
        passed parameters. Valid parameters are:
            name: parts_of_name
                name may include the wild-card (*) character.
            stem: group_stem
            member: member netid
            owner: admin netid
            instructor: instructor netid
                stem="course" will be set when this parameter is passed.
            student: student netid
                stem="course" will be set when this parameter is passed.
            affiliate: affiliate_name
            type: search_type
                Values are 'direct' to search for direct membership and
                'effective' to search for effective memberships. Default is
                direct membership.
            scope: search_scope
                Values are 'one' to limit results to one level of stem name
                and 'all' to return all groups.
        """
        kwargs = dict((k.lower(), v.lower()) for k, v in kwargs.items())
        if 'type' in kwargs and (kwargs['type'] != 'direct' and
                                 kwargs['type'] != 'effective'):
            del(kwargs['type'])

        if 'scope' in kwargs and (kwargs['scope'] != 'one' and
                                  kwargs['scope'] != 'all'):
            del(kwargs['scope'])

        if "instructor" in kwargs or "student" in kwargs:
            kwargs["stem"] = "course"

        url = "/group_sws/v2/search?" + urlencode(kwargs)
        response = self.dao.getURL(url, self._headers({"Accept": "text/xml"}))

        if response.status != 200:
            raise DataFailureException(url, response.status, response.data)

        groups = []
        root = etree.fromstring(response.data)
        e_grs_list = root.find('groupreferences').findall('groupreference')
        for e in e_grs_list:
            group = GroupReference()
            group.uwregid = e.find('regid').text
            group.title = e.find('title').text
            group.description = e.find('description').text
            group.name = e.find('name').text
            groups.append(group)

        return groups

    def get_group_by_id(self, group_id):
        """
        Returns a resttools.Group object for the group identified by the
        passed group ID.
        """
        if not self._is_valid_group_id(group_id):
            raise InvalidGroupID(group_id)

        url = "/group_sws/v2/group/%s" % group_id
        response = self.dao.getURL(url, self._headers({"Accept": "text/xml"}))

        if response.status != 200:
            raise DataFailureException(url, response.status, response.data)

        return self._group_from_xml(response.data)

    def create_group(self, group):
        """
        Creates a group from the passed resttools.Group object.
        """
        body = self._xml_from_group(group)

        url = "/group_sws/v2/group/%s" % group.name
        response = self.dao.putURL(
            url, self._headers({"Accept": "text/xml", "Content-Type": "text/xml"}),
            body)

        if response.status != 201:
            raise DataFailureException(url, response.status, response.data)

        return self._group_from_xml(response.data)

    def put_group(self, group):
        """
        Updates a group from the passed resttools.Group object.
        """
        body = self._xml_from_group(group)

        url = "/group_sws/v2/group/%s" % group.name
        response = self.dao.putURL(
            url, self._headers({"Accept": "text/xml",
                                "Content-Type": "text/xml", "If-Match": "*"}),
            body)

        if response.status != 200:
            raise DataFailureException(url, response.status, response.data)

        return self._group_from_xml(response.data)

    def delete_group(self, group_id):
        """
        Deletes the group identified by the passed group ID.
        """
        if not self._is_valid_group_id(group_id):
            raise InvalidGroupID(group_id)

        url = "/group_sws/v2/group/%s" % group_id
        response = self.dao.deleteURL(url, self._headers({}))

        if response.status != 200:
            raise DataFailureException(url, response.status, response.data)

        return True

    def get_members(self, group_id):
        """
        Returns a list of resttools.GroupMember objects for the group
        identified by the passed group ID.
        """
        if not self._is_valid_group_id(group_id):
            raise InvalidGroupID(group_id)

        url = "/group_sws/v2/group/%s/member" % group_id
        response = self.dao.getURL(url, self._headers({"Accept": "text/xml"}))

        if response.status != 200:
            raise DataFailureException(url, response.status, response.data)

        return self._members_from_xml(response.data)

    def put_membership(self, group_id, members):
        """
        Puts the membership of the group represented by the passed group id.
        Members is an object: name=<member_id>, type=<member_type>
        Returns a list of members not found.
        """
        if not self._is_valid_group_id(group_id):
            raise InvalidGroupID(group_id)

        body = self._xml_from_members(group_id, members)

        url = "/group_sws/v2/group/%s/member" % group_id
        response = self.dao.putURL(
            url, self._headers({"Content-Type": "text/xml", "If-Match": "*"}),
            body)

        if response.status != 200:
            raise DataFailureException(url, response.status, response.data)

        return self._notfoundmembers_from_xml(response.data)

    def put_members(self, group_id, members):
        """
        Puts members into the group represented by the passed group id.
        Members is a list of string. (could be prefaced, e.g. 'u:user_id')
        Returns a list of members not found.
        """
        if not self._is_valid_group_id(group_id):
            raise InvalidGroupID(group_id)

        if len(members) == 0:
            return []

        url = "/group_sws/v2/group/%s/member/%s" % (group_id, ','.join(members))
        response = self.dao.putURL(
            url, self._headers({"Content-Type": "text/xml", "If-Match": "*"}),
            None)

        if response.status != 200:
            raise DataFailureException(url, response.status, response.data)

        return self._notfoundmembers_from_xml(response.data)

    def delete_members(self, group_id, members):
        """
        Delete members from the group represented by the passed group id.
        Members is a list of string.
        Returns a list of members not found.
        """
        if not self._is_valid_group_id(group_id):
            raise InvalidGroupID(group_id)

        if len(members) == 0:
            return True

        url = "/group_sws/v2/group/%s/member/%s" % (group_id, ','.join(members))
        response = self.dao.deleteURL(url, self._headers(
            {"Content-Type": "text/xml", "If-Match": "*"}))

        if response.status != 200:
            raise DataFailureException(url, response.status, response.data)

        return True

    def get_effective_members(self, group_id):
        """
        Returns a list of effective resttools.GroupMember objects for the
        group identified by the passed group ID.
        """
        if not self._is_valid_group_id(group_id):
            raise InvalidGroupID(group_id)

        url = "/group_sws/v2/group/%s/effective_member" % group_id
        response = self.dao.getURL(url, self._headers({"Accept": "text/xml"}))

        if response.status != 200:
            raise DataFailureException(url, response.status, response.data)

        return self._members_from_xml(response.data)

    def get_effective_member_count(self, group_id):
        """
        Returns a count of effective members for the group identified by the
        passed group ID.
        """
        if not self._is_valid_group_id(group_id):
            raise InvalidGroupID(group_id)

        url = "/group_sws/v2/group/%s/effective_member?view=count" % group_id
        response = self.dao.getURL(url, self._headers({"Accept": "text/xml"}))

        if response.status != 200:
            raise DataFailureException(url, response.status, response.data)

        root = etree.fromstring(response.data)
        count = root.find('member_count').get("count")

        return int(count)

    def is_effective_member(self, group_id, netid):
        """
        Returns True if the netid is in the group, False otherwise.
        """
        if not self._is_valid_group_id(group_id):
            raise InvalidGroupID(group_id)

        # GWS doesn't accept EPPNs on effective member checks, for UW users
        netid = re.sub('@washington.edu', '', netid)

        url = "/group_sws/v2/group/%s/effective_member/%s" % (group_id, netid)
        response = self.dao.getURL(url, self._headers({"Accept": "text/xml"}))

        if response.status == 404:
            return False
        elif response.status == 200:
            return True
        else:
            raise DataFailureException(url, response.status, response.data)

    def _group_from_xml(self, data):
        root = etree.fromstring(data)
        gr = root.find('group')
        group_id = gr.find('name').text
        if re.match(r'^course_', group_id):
            group = CourseGroup()
            group.curriculum_abbr = gr.find('course_curr').text.upper()
            group.course_number = gr.find('course_no').text
            group.year = gr.find('course_year').text
            group.quarter = self.QTRS[gr.find('course_qtr').text]
            group.section_id = gr.find('course_sect').text.upper()
            group.sln = gr.find('course_sln').text

            group.instructors = []
            instructors = (gr.find('course_instructors').findall('course_instructor'))
            for instructor in instructors:
                group.instructors.append(GroupMember(name=instructor.text,
                                                     member_type="uwnetid"))
        else:
            group = Group()

        group.name = group_id
        group.uwregid = gr.find('regid').text
        group.title = gr.find('title').text
        group.description = gr.find('description').text
        group.contact = gr.find('contact').text
        group.authnfactor = gr.find('authnfactor').text
        group.classification = gr.find('classification').text
        group.emailenabled = gr.find('emailenabled').text
        group.dependson = gr.find('dependson').text
        group.publishemail = gr.find('publishemail').text

        try:
            group.reporttoorig = gr.find('reporttoorig').text
        except AttributeError:
            # Legacy class name for this attribute
            group.reporttoorig = gr.find('reporttoowner').text

        for user in gr.find('admins').findall('admin'):
            group.admins.append(GroupUser(user.text,
                                          user_type=user.get('type')))

        for user in gr.find('updaters').findall('updater'):
            group.updaters.append(GroupUser(name=user.text,
                                            user_type=user.get("type")))

        for user in gr.find('creators').findall('creator'):
            group.creators.append(GroupUser(name=user.text,
                                            user_type=user.get("type")))

        for user in gr.find('readers').findall('reader'):
            group.readers.append(GroupUser(name=user.text,
                                           user_type=user.get("type")))

        for user in gr.find('optins').findall('optin'):
            group.optins.append(GroupUser(name=user.text,
                                          user_type=user.get("type")))

        for user in gr.find('optouts').findall('optout'):
            group.optouts.append(GroupUser(name=user.text,
                                           user_type=user.get("type")))

        # viewers are not used according to Jim Fox
        # group.viewers = []
        # for user in gr.find('viewers').findall('viewer'):
        #     group.viewers.append(GroupUser(name=user.text,
        #                                    user_type=user.get("type")))
        return group

    def _xml_from_group(self, group):
        template = self._j2env.get_template("group.xml")
        return template.render({"group": group})

    def _members_from_xml(self, data):
        e_mbrs = etree.fromstring(data).find('group').find('members')
        e_mbr_list = e_mbrs.findall('member')

        members = []
        for member in e_mbr_list:
            members.append(GroupMember(name=member.text,
                                       member_type=member.get("type")))

        return members

    def _notfoundmembers_from_xml(self, data):
        members = []
        root = etree.fromstring(data)
        e_nf = root.find('notfoundmembers')
        if e_nf is None:
            return members
        e_nf_list = e_nf.findall('notfoundmember')
        for m in e_nf_list:
            members.append(m.text)

        return members

    def _xml_from_members(self, group_id, members):
        template = self._j2env.get_template("members.xml")
        return template.render({"group_id": group_id, "members": members})

    def _is_valid_group_id(self, group_id):
        if not re.match(r'^[a-z0-9][\w\.-]+$', group_id, re.I):
            return False

        return True

    def _headers(self, headers):
        if self._actas:
            headers = self._add_header(headers, "X-UW-Act-as", self._actas)

        return headers

    def _add_header(self, headers, header, value):
        if not headers:
            return {header: value}

        headers[header] = value
        return headers



"""IRWS service interface"""
import re
import copy
from six.moves.urllib.parse import quote
import json
from resttools.dao import IRWS_DAO
from resttools.models.irws import UWNetId
from resttools.models.irws import Regid
from resttools.models.irws import Subscription
from resttools.models.irws import Person
from resttools.models.irws import Profile
from resttools.models.irws import UWhrPerson
from resttools.models.irws import SdbPerson
from resttools.models.irws import CascadiaPerson
from resttools.models.irws import SccaPerson
from resttools.models.irws import SupplementalPerson
from resttools.models.irws import Pac
from resttools.models.irws import Name
from resttools.models.irws import QnA
from resttools.models.irws import GenericPerson
from resttools.models.irws import PDSEntry

from resttools.exceptions import DataFailureException, InvalidIRWSName
from resttools.exceptions import ResourceNotFound, BadInput

import logging
logger = logging.getLogger(__name__)


class IRWS(object):

    def __init__(self, conf):

        self._service_name = conf['SERVICE_NAME']
        self.dao = IRWS_DAO(conf)

    def _get_code_from_error(self, message):
        try:
            code = int(json.loads(message)['error']['code'])
        except:
            code = (-1)
        return code

    def _clean(self, arg):
        if arg is not None:
            arg = quote(arg)
        return arg

    # v2 - no change
    def get_uwnetid(self, eid=None, regid=None, netid=None, source=None, status=None, ret_array=False):
        """
        Returns an irws.UWNetid object for the given netid or regid.  If the
        netid isn't found, nothing will be returned.  If there is an error
        communicating with the IRWS, a DataFailureException will be thrown.
        if 'all' an array is returned with all the matching netids.
        """
        eid = self._clean(eid)
        regid = self._clean(regid)
        netid = self._clean(netid)

        status_str = ''
        if status is not None:
            status_str = '&status=%d' % status
        if eid is not None and source is not None:
            url = "/%s/v2/uwnetid?validid=%d=%s%s" % (self._service_name, source, eid, status_str)
        elif regid is not None:
            url = "/%s/v2/uwnetid?validid=regid=%s%s" % (self._service_name, regid, status_str)
        elif netid is not None:
            url = "/%s/v2/uwnetid?validid=uwnetid=%s%s" % (self._service_name, netid, status_str)
        else:
            return None
        response = self.dao.getURL(url, {"Accept": "application/json"})

        if response.status == 404:
            return [] if ret_array else None

        if response.status != 200:
            raise DataFailureException(url, response.status, response.data)

        id_data = json.loads(response.data)['uwnetid']
        if ret_array:
            ret = []
            for n in range(0, len(id_data)):
                ret.append(self._uwnetid_from_json_obj(id_data[n]))
            return ret
        else:
            return self._uwnetid_from_json_obj(id_data[0])

    # v2 - no change
    def get_person(self, netid=None, regid=None, eid=None):
        """
        Returns an irws.Person object for the given netid or regid.  If the
        netid isn't found, nothing will be returned.  If there is an error
        communicating with the IRWS, a DataFailureException will be thrown.
        """
        netid = self._clean(netid)
        regid = self._clean(regid)
        eid = self._clean(eid)

        if netid is not None:
            url = "/%s/v2/person?uwnetid=%s" % (self._service_name, netid.lower())
        elif regid is not None:
            url = "/%s/v2/person?validid=regid=%s" % (self._service_name, regid)
        elif eid is not None:
            url = "/%s/v2/person?validid=1=%s" % (self._service_name, eid)
        else:
            return None
        response = self.dao.getURL(url, {"Accept": "application/json"})

        if response.status == 404:
            return None

        if response.status != 200:
            raise DataFailureException(url, response.status, response.data)

        return self._person_from_json(response.data)

    def post_hr_person_by_netid(self, netid, wp_publish=None):
        """
        Update the whitepages publish value for a netid's employee record.
        """
        if wp_publish not in ('Y', 'N', 'E'):
            raise BadInput('Invalid publish option')

        url = self._get_hr_url(netid)
        if url:
            hr_data = {'person': [{'wp_publish': wp_publish}]}
            response = self.dao.postURL(
                url, {"Accept": "application/json"}, json.dumps(hr_data))
            if response.status != 200:
                raise DataFailureException(url, response.status, response.data)
        else:
            raise ResourceNotFound('not an hr person: {}'.format(netid))
        source, eid = url.split('/')[-2:]
        return self.get_uwhr_person(eid, source=source)

    def _get_hr_url(self, netid):
        """
        Given a netid, return the absolute url for a person's employee record
        or None if not an employee.
        """
        person = self.get_person(netid=netid)
        hr_url = None
        if person:
            hr_url = next((url for key, url in person.identifiers.items()
                           if key in ('uwhr', 'hepps')),
                          None)
            if hr_url:
                hr_url = '/{}/v2{}'.format(self._service_name, hr_url)
        return hr_url

    # v2 - no change
    def get_regid(self, netid=None, regid=None):
        """
        Returns an irws.Regid object for the given netid or regid.  If the
        netid isn't found, nothing will be returned.  If there is an error
        communicating with the IRWS, a DataFailureException will be thrown.
        """
        regid = self._clean(regid)
        netid = self._clean(netid)

        if netid is not None:
            url = "/%s/v2/regid?uwnetid=%s" % (self._service_name, netid.lower())
        elif regid is not None:
            url = "/%s/v2/regid?validid=regid=%s" % (self._service_name, regid)
        else:
            return None
        response = self.dao.getURL(url, {"Accept": "application/json"})

        if response.status == 404:
            return None

        if response.status != 200:
            raise DataFailureException(url, response.status, response.data)

        return self._regid_from_json(response.data)

    # v2 - changes
    def get_pw_recover_info(self, netid):
        """
        Returns an irws.Profile object containing password recovery fields
        for the given netid or regid.  If the
        netid isn't found, nothing will be returned.  If there is an error
        communicating with the IRWS, a DataFailureException will be thrown.
        """
        netid = self._clean(netid)

        url = "/%s/v2/profile/validid=uwnetid=%s" % \
            (self._service_name, netid.lower())
        response = self.dao.getURL(url, {"Accept": "application/json"})

        if response.status == 404:
            return None

        if response.status != 200:
            raise DataFailureException(url, response.status, response.data)

        return self._pw_recover_from_json(response.data)

    def put_pw_recover_info(self, netid, profile):
        """
        Updates recover info in netid's profile
        """
        netid = self._clean(netid)

        url = "/%s/v2/profile/validid=uwnetid=%s" % (self._service_name, netid)
        response = self.dao.putURL(url, {"Content-type": "application/json"}, json.dumps(profile.json_data()))

        if response.status >= 500:
            raise DataFailureException(url, response.status, response.data)

        return response.status

    def get_name_by_netid(self, netid):
        """
        Returns a resttools.irws.Name object for the given netid.  If the
        netid isn't found, nothing will be returned.  If there is an error
        communicating with the IRWS, a DataFailureException will be thrown.
        """
        netid = self._clean(netid)

        url = "/%s/v2/name/uwnetid=%s" % (self._service_name, netid.lower())
        response = self.dao.getURL(url, {"Accept": "application/json"})

        if response.status == 404:
            return None

        if response.status != 200:
            raise DataFailureException(url, response.status, response.data)

        return self._name_from_json(response.data)

    def put_name_by_netid(self, netid, first=None, middle=None, last=None):
        name = self.valid_name_json(first=first, middle=middle, last=last)
        url = "/%s/v2/name/uwnetid=%s" % (self._service_name, netid.lower())
        response = self.dao.putURL(
            url, {'Accept': 'application/json'}, name)
        if response.status != 200:
            raise DataFailureException(url, response.status, response.data)
        return response.status

    def valid_name_json(self, first=None, middle=None, last=None):
        """Construct name json to put to IRWS name."""
        if any(not self._valid_name_part(x) for x in (first, middle, last)):
            raise InvalidIRWSName('name too long or has invalid characters')
        if any(not x for x in (first, last)):
            raise InvalidIRWSName('required fields cannot be empty')
        if len(' '.join(x for x in (first, middle, last) if x)) > 80:
            raise InvalidIRWSName(
                'complete display name cannot be longer than 80 characters')

        return json.dumps({'name': [{'preferred_fname': first,
                                     'preferred_mname': middle,
                                     'preferred_sname': last}]})

    @staticmethod
    def _valid_name_part(name):
        regex = r'^[A-Za-z0-9 !$&\'*\-,.?^_`{}~#+%]*$'
        return len(name) <= 64 and re.match(regex, name)

    def get_uwhr_person(self, eid, source='uwhr'):
        """
        Returns an irws.UWhrPerson object for the given eid.
        If the person is not an employee, returns None.
        If the netid isn't found, throws IRWSNotFound.
        If there is an error contacting IRWS, throws DataFailureException.
        """
        eid = self._clean(eid)
        source = self._clean(source)

        url = "/%s/v2/person/%s/%s" % (self._service_name, source, eid)
        response = self.dao.getURL(url, {"Accept": "application/json"})

        if response.status == 404:
            return None

        if response.status != 200:
            raise DataFailureException(url, response.status, response.data)

        person_data = json.loads(response.data)['person'][0]
        return UWhrPerson(**person_data)

    def get_sdb_person(self, sid):
        """
        Returns an irws.SdbPerson object for the given eid.
        If the person is not a student, returns None.
        If the netid isn't found, throws IRWSNotFound.
        If there is an error contacting IRWS, throws DataFailureException.
        """
        sid = self._clean(sid)

        url = "/%s/v2/person/sdb/%s" % (self._service_name, sid)
        response = self.dao.getURL(url, {"Accept": "application/json"})

        if response.status == 404:
            return None

        if response.status != 200:
            raise DataFailureException(url, response.status, response.data)
        person_data = json.loads(response.data)['person'][0]
        return SdbPerson(**person_data)

    def get_cascadia_person(self, id):
        """
        Returns an irws.CascadiaPerson object for the given id.
        If the netid isn't found, throws IRWSNotFound.
        If there is an error contacting IRWS, throws DataFailureException.
        """
        id = self._clean(id)

        url = "/%s/v2/person/cascadia/%s" % (self._service_name, id)
        response = self.dao.getURL(url, {"Accept": "application/json"})

        if response.status == 404:
            return None

        if response.status != 200:
            raise DataFailureException(url, response.status, response.data)

        return self._cascadia_person_from_json(response.data)

    def get_scca_person(self, id):
        """
        Returns an irws.SccaPerson object for the given id.
        If the netid isn't found, throws IRWSNotFound.
        If there is an error contacting IRWS, throws DataFailureException.
        """
        id = self._clean(id)

        url = "/%s/v2/person/scca/%s" % (self._service_name, id)
        response = self.dao.getURL(url, {"Accept": "application/json"})

        if response.status == 404:
            return None

        if response.status != 200:
            raise DataFailureException(url, response.status, response.data)

        return self._scca_person_from_json(response.data)

    def get_supplemental_person(self, id):
        """
        Returns an irws.SupplementalPerson object for the given id.
        If the netid isn't found, throws IRWSNotFound.
        If there is an error contacting IRWS, throws DataFailureException.
        """
        id = self._clean(id)
        url = "/%s/v2/person/supplemental/%s" % (self._service_name, id)
        response = self.dao.getURL(url, {"Accept": "application/json"})

        if response.status == 404:
            return None

        if response.status != 200:
            raise DataFailureException(url, response.status, response.data)

        data = json.loads(response.data)['person'][0]
        return SupplementalPerson(**data)

    def get_generic_person(self, uri):
        """
        Returns an irws.GenericPerson object for the given uri.
        The uris come in from values in irws.Person.identifiers.
        Raises DataFailureExeption on error.
        """
        uri = quote(uri, '/')

        url = '/%s/v2%s' % (self._service_name, uri)
        response = self.dao.getURL(url, {'Accept': 'application/json'})

        if response.status == 404:
            return None
        if response.status != 200:
            raise DataFailureException(url, response.status, response.data)

        return self._generic_person_from_json(response.data)

    def get_subscription(self, netid, subscription):
        """
        Returns an irws.Subscription object for the given netid.  If the
        netid isn't found, nothing will be returned.  If there is an error
        communicating with the IRWS, a DataFailureException will be thrown.
        """
        netid = self._clean(netid)

        url = "/%s/v2/subscription?uwnetid=%s&subscription=%d" % (self._service_name, netid.lower(), subscription)
        response = self.dao.getURL(url, {"Accept": "application/json"})

        if response.status == 404:
            return None

        if response.status != 200:
            raise DataFailureException(url, response.status, response.data)

        return self._subscription_from_json(response.data)

    def get_pdsentry_by_netid(self, netid):
        """
            Returns an irws.PdsEntry object for the given netid.
            If the person doesn't have a pds entry, returns None.
            If the netid isn't found, throws #TODO
            If there is an error contacting IRWS, throws DataFailureException.
            """
        netid = self._clean(netid)

        url = "/%s/v2/pdsentry/validid=uwnetid=%s" % (self._service_name, netid)
        response = self.dao.getURL(url, {"Accept": "application/json"})

        if response.status == 404:
            return None

        if response.status != 200:
            raise DataFailureException(url, response.status, response.data)

        return self._pdsentry_from_json(response.data)

    def put_pac(self, eid, source='uwhr'):
        """
        Creates a PAC for the employee.  Returns the Pac.
        """
        eid = self._clean(eid)
        source = self._clean(source)

        url = "/%s/v2/person/%s/%s/pac?-force" % (self._service_name, source, eid)
        response = self.dao.putURL(url, {"Accept": "application/json"}, '')

        if response.status != 200:
            raise DataFailureException(url, response.status, response.data)

        return self._pac_from_json(response.data)

    def verify_sdb_pac(self, sid, pac):
        """
        Verifies a permanent student PAC. Returns 200 (ok) or 400 (no)
        """
        sid = self._clean(sid)
        pac = self._clean(pac)

        url = "/%s/v2/person/sdb/%s?pac=%s" % (self._service_name, sid, pac)
        response = self.dao.getURL(url, {"Accept": "application/json"})

        if (response.status == 200 or response.status == 400 or response.status == 404):
            return response.status
        raise DataFailureException(url, response.status, response.data)

    def verify_sc_pin(self, netid, pin):
        """
        Verifies a service center one-time pin. Returns 200 (ok) or 400 (no).
        OK clears the pin.
        """
        netid = self._clean(netid)
        pin = self._clean(pin)

        # make sure there is a pin subscription
        url = "/%s/v2/subscription/63/%s" % (self._service_name, netid)
        response = self.dao.getURL(url, {"Accept": "application/json"})
        if response.status == 200:
            sub = json.loads(response.data)['subscription'][0]
            # verify pending subscription and unexpired, unused pac
            if sub['status_code'] != '23' or sub['pac'] != 'Y':
                return 404
        else:
            return response.status

        url = "/%s/v2/subscribe/63/%s?action=1&pac=%s" % (self._service_name, netid, pin)
        response = self.dao.getURL(url, {"Accept": "application/json"})
        if response.status == 200:
            # delete the pac
            url = "/%s/v2/subscribe/63/%s?action=2" % (self._service_name, netid)
            response = self.dao.getURL(url, {"Accept": "application/json"})
            if response.status != 200:
                # the pin was good.  we return OK, but note the error
                logger.error('Delete SC pin failed: %d' % response.status)
            return 200

        if (response.status == 400 or response.status == 404):
            return response.status
        raise DataFailureException(url, response.status, response.data)

    def get_qna(self, netid):
        """
        Returns a list irws.QnA for the given netid.
        """
        netid = self._clean(netid)

        url = "/%s/v2/qna?uwnetid=%s" % (self._service_name, netid)
        response = self.dao.getURL(url, {"Accept": "application/json"})

        if response.status == 404:
            return None

        if response.status != 200:
            raise DataFailureException(url, response.status, response.data)

        return self._qna_from_json(response.data)

    def get_verify_qna(self, netid, answers):
        """
        Verifies that all answers are present and that all are correct.
        answers: ordered list of answers
        """
        questions = self.get_qna(netid)
        if len(questions) != len(answers):
            return False
        for index, answer in enumerate(answers, start=1):
            answer = re.sub(r'\W+', '', answer)
            url = "/%s/v2/qna/%s/%s/check?ans=%s" % (self._service_name, index, quote(netid), quote(answer))
            response = self.dao.getURL(url, {"Accept": "application/json"})
            if response.status in (400, 404):
                logger.debug('qna wrong answer #{}, status = {}'.format(index, response.status))
                return False
            if response.status != 200:
                raise DataFailureException(url, response.status, response.data)
        return True

    def verify_person_attribute(self, netid, attribute, value):
        """
        Verify that the given attribute (eg birthdate) matches the value for the netid.

        Rather than chase all of the person identifier urls client-side, irws will return the
        list of identifiers. For birthdate, IRWS has the added value of discarding silly
        birthdates and matching on partial birthdates.
        """
        netid = self._clean(netid)
        attribute = self._clean(attribute)
        value = self._clean(value)

        url = "/%s/v2/person?uwnetid=%s&%s=%s" % (self._service_name, netid, attribute, value)
        return self.dao.getURL(url, {'Accept': 'application/json'}).status == 200

    def _cascadia_person_from_json(self, data):
        """
        Internal method, for creating the CascadiaPerson object.
        """
        person_data = json.loads(data)['person'][0]
        person = CascadiaPerson()
        person.validid = person_data['validid']
        person.regid = person_data['regid']
        person.lname = person_data['lname']
        person.categories = person_data['categories']
        if 'birthdate' in person_data:
            person.birthdate = person_data['birthdate']
        if 'department' in person_data:
            person.department = person_data['department']
        if 'in_feed' in person_data:
            person.in_feed = person_data['in_feed']
        person.status_code = person_data.get('status_code', person.__class__.status_code)
        return person

    def _scca_person_from_json(self, data):
        """
        Internal method, for creating the SccaPerson object.
        """
        person_data = json.loads(data)['person'][0]
        person = SccaPerson()
        person.validid = person_data['validid']
        person.regid = person_data['regid']
        person.lname = person_data['lname']
        person.categories = person_data['categories']
        if 'birthdate' in person_data:
            person.birthdate = person_data['birthdate']
        if 'scca_company' in person_data:
            person.scca_company = person_data['scca_company']
        if 'scca_cca_eppn' in person_data:
            person.scca_cca_eppn = person_data['scca_cca_eppn']
        if 'scca_fhc_eppn' in person_data:
            person.scca_fhc_eppn = person_data['scca_fhc_eppn']
        if 'in_feed' in person_data:
            person.in_feed = person_data['in_feed']
        person.status_code = person_data.get('status_code', person.__class__.status_code)
        return person

    def _person_from_json(self, data):
        persj = json.loads(data)['person'][0]
        idj = persj['identity']
        person = Person()
        person.regid = idj['regid']
        if 'lname' in idj:
            person.lname = idj['lname']
        if 'fname' in idj:
            person.fname = idj['fname']
        if 'identifiers' in idj:
            person.identifiers = copy.deepcopy(idj['identifiers'])
        return person

    def _regid_from_json(self, data):
        rj = json.loads(data)['regid'][0]
        regid = Regid()
        regid.regid = rj['regid']
        regid.entity_code = rj['entity_code']
        regid.entity_name = rj['entity_name']
        regid.status_code = rj['status_code']
        regid.status_name = rj['status_name']
        return regid

    def _pw_recover_from_json(self, data):
        info = json.loads(data)['profile'][0]
        ret = Profile()
        if 'validid' in info:
            ret.validid = info['validid']
        if 'recover_contacts' in info:
            ret.recover_contacts = info['recover_contacts']
        if 'recover_block_reasons' in info:
            ret.recover_block_reasons = info['recover_block_reasons']
        return ret

    def _uwnetid_from_json_obj(self, id_data):
        uwnetid = UWNetId()
        uwnetid.uwnetid = id_data['uwnetid']
        uwnetid.validid = id_data['validid']
        uwnetid.uid = id_data['uid']
        return uwnetid

    def _subscription_from_json(self, data):
        sub_data = json.loads(data)['subscription'][0]
        subscription = Subscription()
        subscription.uwnetid = sub_data['uwnetid']
        subscription.subscription_code = sub_data['subscription_code']
        subscription.subscription_name = sub_data['subscription_name']
        subscription.subscription_data = sub_data.get('subscription_data', None)
        return subscription

    def _pdsentry_from_json(self, data):
        """
            Internal method, for creating the PDSEntry object.
            """
        person_data = json.loads(data)['pdsentry'][0]['entry']
        person = PDSEntry()
        person.regid = person_data.get('uwRegID', '')
        person.objectclass = person_data.get('objectClass', [])
        person.test = person_data.get('uwTest', '')
        person.eid = person_data.get('uwEmployeeID', '')
        person.ewpname = person_data.get('uwEWPName', '')
        person.ewpdept = person_data.get('uwEWPDept', [])
        person.ewpemail = person_data.get('uwEWPEmail', [])
        person.ewpphone = person_data.get('uwEWPPhone', [])
        person.ewptitle = person_data.get('uwEWPTitle', [])
        person.ewpaddr = person_data.get('uwEWPAddr', [])
        person.ewppublish = person_data.get('uwEWPPublish', '')
        person.employeehomedept = person_data.get('uwEmployeeHomeDepartment', '')
        person.employeemailstop = person_data.get('uwEmployeeMailstop', '')
        person.studentsystemkey = person_data.get('uwStudentSystemKey', '')
        person.sid = person_data.get('uwStudentID', '')
        person.swpname = person_data.get('uwSWPName', '')
        person.swpdept = person_data.get('uwSWPDept', [])
        person.swpemail = person_data.get('uwSWPEmail', '')
        person.swpclass = person_data.get('uwSWPClass', '')
        person.swppublish = person_data.get('uwSWPPublish', '')
        person.developmentid = person_data.get('uwDevelopmentID', '')
        person.wppublish = person_data.get('uwWPPublish', '')
        person.displayname = person_data.get('displayName', '')
        person.registeredname = person_data.get('uwPersonRegisteredName', '')
        person.registeredfirstmiddle = person_data.get('uwPersonRegisteredFirstMiddle', '')
        person.registeredsurname = person_data.get('uwPersonRegisteredSurname', '')
        person.cn = person_data.get('cn', '')
        person.sn = person_data.get('sn', '')
        person.preferredname = person_data.get('uwPersonPreferredName', '')
        person.preferredfirst = person_data.get('uwPersonPreferredFirst', '')
        person.preferredsurname = person_data.get('uwPersonPreferredSurname', '')
        person.uwnetid = person_data.get('uwNetID', '')
        person.uidnumber = person_data.get('uidNumber', '')
        person.edupersonaffiliation = person_data.get('eduPersonAffiliation', '')
        return person

    def _pac_from_json(self, data):
        pac_data = json.loads(data)['person'][0]
        pac = Pac()
        pac.pac = pac_data['pac']
        pac.expiration = pac_data['expiration']
        return pac

    def _name_from_json(self, data):
        nd = json.loads(data)['name'][0]
        name = Name()
        name.validid = nd['validid']
        if 'formal_cname' in nd:
            name.formal_cname = nd['formal_cname']
        if 'formal_fname' in nd:
            name.formal_fname = nd['formal_fname']
        if 'formal_sname' in nd:
            name.formal_lname = nd['formal_sname']
        if 'formal_privacy' in nd:
            name.formal_privacy = nd['formal_privacy']
        if 'display_cname' in nd:
            name.display_cname = nd['display_cname']
        if 'display_fname' in nd:
            name.display_fname = nd['display_fname']
        if 'display_mname' in nd:
            name.display_mname = nd['display_mname']
        if 'display_sname' in nd:
            name.display_lname = nd['display_sname']
        if 'display_privacy' in nd:
            name.display_privacy = nd['display_privacy']
        if 'preferred_cname' in nd:
            name.preferred_cname = nd['preferred_cname']
        if 'preferred_fname' in nd:
            name.preferred_fname = nd['preferred_fname']
        if 'preferred_mname' in nd:
            name.preferred_mname = nd['preferred_mname']
        if 'preferred_sname' in nd:
            name.preferred_lname = nd['preferred_sname']
        if 'preferred_privacy' in nd:
            name.preferred_privacy = nd['preferred_privacy']
        return name

    def _qna_from_json(self, data):
        q_list = json.loads(data)['qna']
        ret = []
        for q in q_list:
            qna = QnA()
            qna.uwnetid = q['uwnetid']
            qna.ordinal = q['ordinal']
            qna.question = q['question']
            # qna.answer = q['answer']
            ret.append(qna)
        return ret

    def _generic_person_from_json(self, data):
        """
        Internal method to create a GenericPerson object.
        """
        person_data = json.loads(data)['person'][0]
        person = GenericPerson()
        attributes = [attribute for attribute in dir(GenericPerson) if not attribute.startswith('_')]
        for attribute in attributes:
            # set the attribute to the value in person_data, or if not set there,
            # the default attribute value
            setattr(person, attribute, person_data.get(attribute, getattr(GenericPerson, attribute)))
        return person



import json
from nose.tools import *
from mock import patch

from resttools.nws import NWS

import resttools.test.test_settings as settings
import logging.config
logging.config.dictConfig(settings.LOGGING)
logger = logging.getLogger(__name__)


class NWS_Test():

    def __init__(self):
        self.nws = NWS(settings.NWS_CONF)

    def test_get_netid_admins(self):
        admins = self.nws.get_netid_admins('groups')
        eq_(len(admins), 3)
        for a in admins:
            ok_((a.name == 'fox' and a.type == 'netid') or
                (a.name == 'dors' and a.type == 'netid') or
                (a.name == 'u_fox_browser6' and a.type == 'group'))

    def test_get_netid_admins_no_list(self):
        with patch('resttools.nws.NWS_DAO') as nws_dao:
            mock_response = nws_dao.return_value.getURL.return_value
            mock_response.status = 200
            mock_response.data = json.dumps({"timeStamp": "now", "uwNetID": "joe"})
            nws = NWS(settings.NWS_CONF)
            eq_(nws.get_netid_admins('joe'), [])

    def test_get_netid_pw(self):
        pw = self.nws.get_netid_pwinfo('groups')
        eq_(pw.min_len, 8)
        eq_(pw.kerb_status, 'Active')

    def test_pwinfo_from_json(self):
        pw = self.nws._pwinfo_from_json(json.dumps({'minimumLength': 50000}))
        eq_(pw.__dict__, {'min_len': 50000, 'kerb_status': 'Active', 'last_change': ''})

        pw = self.nws._pwinfo_from_json(json.dumps(
                                        {'minimumLength': 4000, 'lastChange': 'tomorrow', 'kerbStatus': 'dissed'}))
        eq_(pw.__dict__, {'min_len': 4000, 'kerb_status': 'dissed', 'last_change': 'tomorrow'})

        pw = self.nws._pwinfo_from_json(json.dumps({}))
        eq_(pw, None)



# NWS representations
class UWNetIdAdmin():
    name = ''
    role = ''
    type = 'netid'

    def __init__(self, name='', role='', type='netid'):
        self.name = name
        self.role = role
        self.type = type


class UWNetIdPwInfo():
    min_len = 0
    last_change = ''
    kerb_status = 'Active'

    def __init__(self, min_len=None, last_change=None, kerb_status=None):
        self.min_len = min_len if min_len is not None else UWNetIdPwInfo.min_len
        self.last_change = last_change if last_change is not None else UWNetIdPwInfo.last_change
        self.kerb_status = kerb_status if kerb_status is not None else UWNetIdPwInfo.kerb_status




class GroupReference():
    uwregid = ''
    name = ''
    title = ''
    description = ''
    url = ''

    def __str__(self):
        return "{uwregid: %s, name: %s, title: %s, description: %s}" % (
            self.uwregid, self.name, self.title, self.description)


class Group():
    CLASSIFICATION_NONE = "u"
    CLASSIFICATION_PUBLIC = "p"
    CLASSIFICATION_RESTRICTED = "r"
    CLASSIFICATION_CONFIDENTIAL = "c"

    CLASSIFICATION_TYPES = (
        (CLASSIFICATION_NONE, "Unclassified"),
        (CLASSIFICATION_PUBLIC, "Public"),
        (CLASSIFICATION_RESTRICTED, "Restricted"),
        (CLASSIFICATION_CONFIDENTIAL, "Confidential")
    )

    uwregid = ''

    name = ''
    title = ''
    description = ''
    contact = ''
    authnfactor = 1
    classification = ''
    emailenabled = ''
    dependson = ''
    publishemail = ''
    reporttoorig = ''

    def __init__(self, *args, **kwargs):
        self.admins = []
        self.creators = []
        self.optins = []
        self.optouts = []
        self.readers = []
        self.updaters = []

    def __str__(self):
        return "{uwregid: %s, name: %s, title: %s, description: %s}" % (
            self.uwregid, self.name, self.title, self.description)

    def has_regid(self):
        return self.uwregid is not None and len(self.uwregid) == 32


class CourseGroup(Group):
    SPRING = "spring"
    SUMMER = "summer"
    AUTUMN = "autumn"
    WINTER = "winter"

    QUARTERNAME_CHOICES = (
        (SPRING, "Spring"),
        (SUMMER, "Summer"),
        (AUTUMN, "Autumn"),
        (WINTER, "Winter"),
    )

    curriculum_abbr = ''
    course_number = ''
    year = 0
    quarter = ''
    section_id = ''
    sln = 0


class GroupUser():
    UWNETID_TYPE = "uwnetid"
    EPPN_TYPE = "eppn"
    GROUP_TYPE = "group"
    DNS_TYPE = "dns"
    NONE_TYPE = "none"

    TYPE_CHOICES = (
        (UWNETID_TYPE, "UWNetID"),
        (EPPN_TYPE, "ePPN"),
        (GROUP_TYPE, "Group ID"),
        (DNS_TYPE, "Hostname"),
        (NONE_TYPE, "Anyone"),
    )

    def __init__(self, name='', user_type=''):
        self.name = name
        self.user_type = user_type

    def is_uwnetid(self):
        return self.user_type == self.UWNETID_TYPE

    def is_eppn(self):
        return self.user_type == self.EPPN_TYPE

    def is_group(self):
        return self.user_type == self.GROUP_TYPE

    def is_none(self):
        return self.user_type == self.NONE_TYPE

    def __eq__(self, other):
        return self.name == other.name and self.user_type == other.user_type

    def __str__(self):
        return "{name: %s, user_type: %s}" % (
            self.name, self.user_type)


class GroupMember():
    UWNETID_TYPE = "uwnetid"
    EPPN_TYPE = "eppn"
    GROUP_TYPE = "group"
    DNS_TYPE = "dns"

    TYPE_CHOICES = (
        (UWNETID_TYPE, "UWNetID"),
        (EPPN_TYPE, "ePPN"),
        (GROUP_TYPE, "Group ID"),
        (DNS_TYPE, "Hostname"),
    )

    def __init__(self, name='', member_type=''):
        self.name = name
        self.member_type = member_type

    def is_uwnetid(self):
        return self.member_type == self.UWNETID_TYPE

    def is_eppn(self):
        return self.member_type == self.EPPN_TYPE

    def is_group(self):
        return self.member_type == self.GROUP_TYPE

    def __eq__(self, other):
        return self.name == other.name and self.member_type == other.member_type

    def __str__(self):
        return "{name: %s, user_type: %s}" % (
            self.name, self.member_type)



from base64 import b64encode, b64decode
from datetime import datetime


# IRWS Name
class Name():
    validid = ''
    formal_cname = ''
    formal_fname = ''
    formal_lname = ''
    formal_privacy = ''
    display_cname = ''
    display_fname = ''
    display_mname = ''
    display_lname = ''
    display_privacy = ''
    preferred_cname = ''
    preferred_fname = ''
    preferred_mname = ''
    preferred_lname = ''
    preferred_privacy = ''

    def json_data(self):
        return {"formal_cname": self.formal_cname,
                "formal_fname": self.formal_fname,
                "formal_lname": self.formal_lname,
                "formal_privacy": self.formal_privacy,
                "display_cname": self.display_cname,
                "display_fname": self.display_fname,
                "display_mname": self.display_mname,
                "display_lname": self.display_lname,
                "display_privacy": self.display_privacy,
                "preferred_cname": self.preferred_cname,
                "preferred_fname": self.preferred_fname,
                "preferred_mname": self.preferred_mname,
                "preferred_lname": self.preferred_lname,
                "preferred_privacy": self.preferred_privacy,
                }


# IRWS Profile (only the recover data for now)
class Profile():
    validid = ''
    recover_block_reasons = []
    recover_contacts = []

    def json_data(self):
        return {
            'profile': [
                {"recover_contacts": self.recover_contacts,
                 "recover_block_reasons": self.recover_block_reasons
                 }
            ]}


# IRWS Person
class Person():
    regid = ''
    lname = ''
    fname = ''
    identifiers = {}


class UWhrPerson(object):
    """IRWS UWhr Person"""
    validid = ''
    regid = ''
    studentid = ''
    birthdate = ''

    fname = ''
    lname = ''
    categories = ()
    contact_email = ()
    workday_home_email = ''
    org_supervisor = ''

    wp_name = ''
    wp_lname = ''
    wp_fname = ''
    wp_department = ()
    wp_email = ()
    wp_phone = ()
    wp_title = ()
    wp_address = ()
    wp_mobile = ()
    wp_facsimile = ()
    wp_pager = ()
    wp_touchdial = ()  # TTY/TDD
    wp_voicemail = ()
    wp_publish = 'N'  # hepps-only attribute
    wp_publish_options = None

    college = ''
    department = ''
    home_department = ''
    mailstop = ''
    unit = ''

    emp_ecs_code = ''
    emp_status_code = ''
    workday_last_active = ''
    workday_emp_status = ''
    workday_con_status = ''
    workday_aff_status = ''

    budget = ''
    faccode = ''
    source_code = ''
    source_name = ''
    status_code = ''
    status_name = ''

    pac = ''

    in_feed = ''

    created = ''
    updated = ''

    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)

        # uwhr doesn't use these. They can come out when no clients use it.
        self.wp_fname, self.wp_lname = self.fname, self.wp_lname
        self.wp_name = ' '.join([self.fname, self.lname])


# IRWS Sdb Person
class SdbPerson(object):
    validid = ''
    regid = ''
    studentid = ''
    birthdate = ''

    fname = ''
    lname = ''
    categories = ()

    branch = ''
    college = ''
    department = ''

    source_code = ''
    source_name = ''
    # status: 1=active, 3=former
    status_code = ''
    status_name = ''
    pac = ''

    cnet_id = ''
    cnet_user = ''

    wp_publish = 'N'
    wp_publish_options = None
    wp_title = ()
    wp_phone = ()
    wp_department = ()
    wp_email = ()

    in_feed = ''
    created = ''
    updated = ''

    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)


# IRWS Supplemental Person
class SupplementalPerson():
    validid = ''
    regid = ''
    lname = ''
    categories = []
    contact_email = []
    comment_code = ''
    comment_name = ''
    sponsor_id = ''
    college = ''
    source_code = ''
    source_name = ''
    status_code = ''
    status_name = ''
    in_feed = ''
    id_proofing = {}
    created = ''
    updated = ''

    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)


# IRWS Cascadia person
class CascadiaPerson():
    validid = ''
    regid = ''
    categories = []
    lname = ''
    fname = ''
    birthdate = ''
    department = ''
    in_feed = ''
    status_code = ''


# IRWS Scca person
class SccaPerson():
    validid = ''
    regid = ''
    categories = []
    lname = ''
    fname = ''
    birthdate = ''
    scca_company = ''
    scca_cca_eppn = ''
    scca_fhc_eppn = ''
    in_feed = ''
    status_code = ''


# IRWS GenericPerson
class GenericPerson():
    validid = ''
    regid = ''
    lname = ''
    fname = ''
    contact_email = []
    categories = []
    source_code = ''
    status_code = ''


# IRWS UWNetId
class UWNetId():
    uwnetid = ''
    accid = ''
    validid = ''
    uid = ''
    luid = ''
    disenfran = ''
    netid_code = ''
    netid_name = ''
    status_code = ''
    status_name = ''
    logname = ''
    created = ''
    updated = ''


# IRWS Regid
class Regid():
    regid = ''
    entity_code = ''
    entity_name = ''
    status_code = ''
    status_name = ''

    created = ''
    updated = ''


# IRWS Subscription
class Subscription():
    uwnetid = ''
    subscription_code = ''
    subscription_name = ''
    subscription_data = ''
    notify_code = ''
    status_code = ''
    status_name = ''
    logname = ''
    created = ''
    updated = ''


# IRWS PAC
class Pac():
    pac = ''
    expiration = ''


# IRWS QnA
class QnA():
    uwnetid = ''
    ordinal = ''
    question = ''
    answer = ''


# IRWS PDSEntry
class PDSEntry():
    regid = ''
    objectclass = []
    test = ''
    eid = ''
    ewpname = ''
    ewpdept = []
    ewpemail = []
    ewpphone = []
    ewptitle = []
    ewpaddr = []
    ewppublish = ''
    employeehomedepartment = ''
    employeemailstop = ''
    studentsystemkey = ''
    sid = ''
    swpname = ''
    swpdept = []
    swpemail = ''
    swpclass = ''
    swppublish = ''
    developmentid = ''
    wppublish = ''
    displayname = ''
    registeredname = ''
    registeredfirstmiddle = ''
    registeredsurname = ''
    cn = ''
    sn = ''
    preferredname = ''
    preferredfirst = ''
    preferredsurname = ''
    uwnetid = ''
    uidnumber = ''
    edupersonaffiliation = []



"""
Contains objects used by the non-HTTP DAO implementations
"""


class MockHTTP(object):
    """
    An alternate object to HTTPResponse, for non-HTTP DAO
    implementations to use.  Implements the API of HTTPResponse
    as needed.
    """
    status = 0
    data = ""
    headers = {}

    def read(self):
        """
        Returns the document body of the request.
        """
        return self.data

    def getheader(self, field, default=''):
        """
        Returns the HTTP response header field, case insensitively
        """
        if self.headers:
            for header in self.headers:
                if field.lower() == header.lower():
                    return self.headers[header]

        return default



"""
Contains the custom exceptions used by the restclients.
"""


class PhoneNumberRequired(Exception):
    """Exception for missing phone number."""
    pass


class InvalidPhoneNumber(Exception):
    """Exception for invalid phone numbers."""
    pass


class InvalidNetID(Exception):
    """Exception for invalid netid."""
    pass


class InvalidRegID(Exception):
    """Exception for invalid regid."""
    pass


class InvalidEmployeeID(Exception):
    """Exception for invalid employee id."""
    pass


class InvalidUUID(Exception):
    """Exception for invalid UUID."""
    pass


class InvalidSectionID(Exception):
    """Exception for invalid section id."""
    pass


class InvalidSectionURL(Exception):
    """Exception for invalid section url."""
    pass


class InvalidGroupID(Exception):
    """Exception for invalid group id."""
    pass


class InvalidIdCardPhotoSize(Exception):
    """Exception for invalid photo size."""
    pass


class InvalidEndpointProtocol(Exception):
    """Exception for invalid endpoint protocol."""
    pass


class InvalidCanvasIndependentStudyCourse(Exception):
    """Exception for invalid Canvas course."""
    pass


class InvalidCanvasSection(Exception):
    """Exception for invalid Canvas section."""
    pass


class InvalidGradebookID:
    """Exception for invalid gradebook id."""
    pass


class InvalidIRWSName(Exception):
    """Exception for invalid IRWS name."""
    pass


class ResourceNotFound(LookupError):
    """Exception when a resource doesn't exist."""
    pass


class BadInput(ValueError):
    """Invalid argument or payload."""
    pass


class DataFailureException(Exception):
    """
    This exception means there was an error fetching content
    in one of the rest clients.  You can get the url that failed
    with .url, the status of the error with .status, and any
    message with .msg
    """
    def __init__(self, url, status, msg):
        self.url = url
        self.status = status
        self.msg = msg

    def __str__(self):
        return ("Error fetching %s.  Status code: %s.  Message: %s." %
                (self.url, self.status, self.msg))



import json
from nose.tools import *
from resttools.irws import IRWS
from resttools.exceptions import (InvalidIRWSName, DataFailureException,
                                  BadInput, ResourceNotFound)
from resttools.models.irws import SdbPerson
import resttools.test.test_settings as settings
import logging.config
logging.config.dictConfig(settings.LOGGING)
logger = logging.getLogger(__name__)


class IRWS_Test():

    def __init__(self):
        self.irws = IRWS(settings.IRWS_CONF)

    def test_get_name_by_netid(self):
        name = self.irws.get_name_by_netid('javerage')
        eq_(name.preferred_cname, 'JAMES AVERAGE STUDENT-P')
        eq_(name.preferred_privacy, 'Public')

    def test_get_uwnetid_by_regid(self):
        netid = self.irws.get_uwnetid(regid='DC5C0C166A7C11D5A4AE0004AC494FFE', status=30)
        eq_(netid.uwnetid, 'lucy123wf')
        eq_(netid.uid, '2024')
        netids = self.irws.get_uwnetid(regid='DC5C0C166A7C11D5A4AE0004AC494FFE', status=30, ret_array=True)
        eq_(len(netids), 2)
        eq_(netids[1].uwnetid, 'junk4')

    def test_get_uwnetid_by_eid(self):
        netid = self.irws.get_uwnetid(eid='003433412', source=1, status=30)
        eq_(netid.uwnetid, 'fox')
        eq_(netid.uid, '994010')

    def test_get_person_by_netid(self):
        person = self.irws.get_person(netid='wdspud867')
        eq_(person.lname, 'Daywork')
        eq_(person.fname, 'Spud')
        eq_(person.identifiers['hepps'], '/person/hepps/867003233')

    def test_get_person_by_netid_nonexists(self):
        person = self.irws.get_person(netid='pud867')
        eq_(person, None)

    def test_post_hr_person_by_netid(self):
        uwhr = self.irws.post_hr_person_by_netid('javerage', wp_publish='E')
        eq_(uwhr.wp_publish, 'E')

    def test_post_hr_person_bad_publish_value(self):
        assert_raises(BadInput,
                      self.irws.post_hr_person_by_netid,
                      'javerage', wp_publish='X')

    def test_post_hr_person_not_employee(self):
        assert_raises(ResourceNotFound,
                      self.irws.post_hr_person_by_netid,
                      'notaperson', wp_publish='Y')

    def test_get_uwhr_person_hepps(self):
        uwhr = self.irws.get_uwhr_person('123456789', source='hepps')
        eq_(uwhr.lname, 'STUDENT')

    def test_get_sdb_person(self):
        sdb = self.irws.get_sdb_person('000083856')
        eq_(sdb.lname, 'STUDENT')

    def test_get_sdb_person_default_wp_publish(self):
        data = dict(
            validid='123', regid='000', studentid='111', fname='F', lname='L',
            categories=[], source_code='1', source_name='N', status_code='2',
            status_name='M')
        sdb = SdbPerson(**data)
        eq_(sdb.wp_publish_options, None)
        eq_(sdb.validid, '123')

    def test_get_supplemental_person(self):
        s = self.irws.get_supplemental_person('88E13ABD')
        eq_(s.lname, 'AVERAGE')

    def test_get_regid(self):
        r = self.irws.get_regid(netid='joeuser')
        eq_(r.entity_code, '10')
        eq_(r.status_code, '50')
        r = self.irws.get_regid(regid='FC8E9A4FD5A940ACAC6306EA7DC7D742')
        eq_(r.entity_code, '10')
        eq_(r.status_code, '50')

    def test_get_qna(self):
        qna = self.irws.get_qna('user1q')
        eq_(len(qna), 3)
        eq_(qna[0].ordinal, '1')
        eq_(qna[1].ordinal, '2')
        eq_(qna[2].ordinal, '3')

    def test_verify_qna(self):
        correct = ['skyblue', 'mememe', 'begood']
        incorrect = ['skyblue', 'NOTmememe', 'begood']
        eq_(self.irws.get_verify_qna('user1q', correct), True)
        eq_(self.irws.get_verify_qna('user1q', incorrect), False)
        eq_(self.irws.get_verify_qna('user1q', correct[:2]), False)

    def test_verify_person_attibute(self):
        st = self.irws.verify_person_attribute('javerage', 'birthdate', '2015-10-04')
        eq_(st, True)
        st = self.irws.verify_person_attribute('javerage', 'birthdate', '2015-10-03')
        eq_(st, False)

    def test_get_cascadia_person(self):
        g = self.irws.get_cascadia_person('960031000')
        eq_(g.lname, 'CASLAST')
        eq_(g.birthdate, '1981-01-01')
        eq_(g.department, 'STU')
        eq_(g.status_code, '3')

    def test_get_scca_person(self):
        g = self.irws.get_scca_person('FHSC40999')
        eq_(g.lname, 'CCALAST')
        eq_(g.scca_company, 'CCA')
        eq_(g.scca_cca_eppn, 'ccalast@seattlecca.org')
        eq_(g.status_code, '3')

    def test_get_generic_person(self):
        g = self.irws.get_generic_person('/person/generic/01234')
        eq_(g.lname, 'LEGACYEMAIL')
        eq_(g.contact_email, ['legacyemail@example.com'])
        eq_(g.source_code, '2')
        # not set in our payload, assert the default value
        eq_(g.status_code, '')

    def test_verify_sc_pin(self):
        st = self.irws.verify_sc_pin('user1s', '321abc')
        eq_(st, 404)
        st = self.irws.verify_sc_pin('user1s', '123456')
        eq_(st, 200)
        st = self.irws.verify_sc_pin('user1q', '123456')
        eq_(st, 404)

    def test_put_name_by_netid(self):
        # prime the cache first
        # irws.get_name_by_netid('javerage')

        response = self.irws.put_name_by_netid(
            'javerage', first='J', middle='', last='Student')
        eq_(200, response)
        name = self.irws.get_name_by_netid('javerage')
        eq_('J', name.preferred_fname)
        eq_('', name.preferred_mname)
        eq_('Student', name.preferred_lname)

    def test_put_name_by_netid_no_user(self):
        assert_raises(DataFailureException,
                      self.irws.put_name_by_netid,
                      'nonuser',
                      first='J', middle='', last='Student')

    def test_valid_name_part_good(self):
        ok_(self.irws._valid_name_part('james'))
        ok_(self.irws._valid_name_part(' '))
        ok_(self.irws._valid_name_part(' !$&\'*-,.?^_`{}~#+%'))

    def test_valid_name_part_bad(self):
        bad_chars = '"():;<>[\]|@'
        assert not self.irws._valid_name_part(u'Jos\xe9')  # utf-8
        for c in bad_chars:
            assert not self.irws._valid_name_part(c)

    def test_valid_name_part_too_long(self):
        # 64 is the magic number
        bad_name = 'a' * 65
        assert not self.irws._valid_name_part(bad_name)
        # one less should be good
        assert self.irws._valid_name_part(bad_name[:-1])

    def test_valid_name_json_good(self):
        names = json.loads(
            self.irws.valid_name_json(first='joe', middle='average', last='user'))
        name = names['name'][0]
        eq_(name['preferred_fname'], 'joe')
        eq_(name['preferred_mname'], 'average')
        eq_(name['preferred_sname'], 'user')

    def test_valid_irws_name_empty_middle_name(self):
        names = json.loads(self.irws.valid_name_json(first='joe', middle='', last='user'))
        eq_(names['name'][0]['preferred_mname'], '')

    def test_valid_name_json_required_fields_missing(self):
        bad_data_list = [
            dict(first='', middle='average', last='user'),
            dict(first='joe', middle='average', last='')]

        for bad_data in bad_data_list:
            assert_raises(InvalidIRWSName,
                          self.irws.valid_name_json,
                          **bad_data)

    def test_valid_irws_name_bad_characters(self):
        bad_data_list = [
            ('@', 'average', 'user'),
            ('joe', '@', 'user'),
            ('joe', 'average', '@'),
            ]

        for bad_data in bad_data_list:
            bad_name = dict(first=bad_data[0], middle=bad_data[1], last=bad_data[2])
            assert_raises(InvalidIRWSName,
                          self.irws.valid_name_json,
                          **bad_name)

    def test_valid_irws_name_too_long(self):
        # 80 is the magic number
        bad_data_list = [
            ('f' * 30, 'm' * 30, 'l' * 19),
            ('f' * 40, '', 'l' * 40)]

        for bad_data in bad_data_list:
            # one less character will pass
            good_name = (bad_data[0:2] + (bad_data[2][:-1],))
            ok_(self.irws.valid_name_json(first=good_name[0],
                                          middle=good_name[1], last=good_name[2]))
            # failure
            assert_raises(InvalidIRWSName,
                          self.irws.valid_name_json,
                          first=bad_data[0], middle=bad_data[1], last=bad_data[2])



"""
A centralized the mock data access
"""
import sys
import os
from os.path import abspath, dirname
import re
import json
import logging
from resttools.mock.mock_http import MockHTTP
fs_encoding = sys.getfilesystemencoding() or sys.getdefaultencoding()
app_resource_dirs = []


def get_mockdata_url(service_name, conf,
                     url, headers):
    """
    :param service_name:
        possible "irws", "pws", "gws", etc.
    """
    dir_base = dirname(__file__)
    app_root = abspath(dir_base)
    response = _load_resource_from_path(app_root, service_name, conf, url, headers)
    if response:
        return response

    # If no response has been found in any installed app, return a 404
    logger = logging.getLogger(__name__)
    logger.debug("404 for url %s")
    response = MockHTTP()
    response.status = 404
    return response


def _load_resource_from_path(app_root, service_name, conf, url, headers):

    logger = logging.getLogger(__name__)
    mock_root = os.path.realpath(os.path.join(app_root, '../mock'))
    std_root = os.path.join(mock_root, service_name)
    if 'MOCK_ROOT' in conf and conf['MOCK_ROOT'] is not None:
        mock_root = conf['MOCK_ROOT']
    root = os.path.join(mock_root, service_name)

    if url == "///":
        # Just a placeholder to put everything else in an else.
        # If there are things that need dynamic work, they'd go here
        pass
    else:
        try:
            file_path = convert_to_platform_safe(root + url)
            logger.debug('try1: ' + file_path)
            if os.path.isdir(file_path):
                file_path = file_path + '.resource'
            handle = open(file_path)
        except IOError:
            if std_root is not mock_root:
                try:
                    file_path = convert_to_platform_safe(std_root + url)
                    logger.debug('try2: ' + file_path)
                    if os.path.isdir(file_path):
                        file_path = file_path + '.resource'
                    handle = open(file_path)
                except IOError:
                    return

        logger.debug("URL: %s; File: %s" % (url, file_path))

        response = MockHTTP()
        response.status = 200
        data = handle.read()
        cut = data.find('MOCKDATA-MOCKDATA-MOCKDATA')
        if cut >= 0:
            data = data[(data.find('\n', cut)+1):]
        response.data = data
        response.headers = {"X-Data-Source": service_name + " file mock data", }

        try:
            headers = open(handle.name + '.http-headers')
            data = headers.read()
            cut = data.find('MOCKDATA-MOCKDATA-MOCKDATA')
            if cut >= 0:
                data = data[(data.find('\n', cut) + 1):]
            file_values = json.loads(data)

            if "headers" in file_values:
                response.headers = dict(response.headers.items() + file_values['headers'].items())

            if 'status' in file_values:
                response.status = file_values['status']

            else:
                response.headers = dict(response.headers.items() + file_values.items())

        except IOError:
            pass

        return response


def post_mockdata_url(service_name, conf, url, headers, body, dir_base=dirname(__file__)):
    """
    :param service_name:
        possible "sws", "pws", "book", "hfs", etc.
    """
    # Currently this post method does not return a response body
    response = MockHTTP()
    if body is not None:
        if "dispatch" in url:
            response.status = 200
        else:
            response.status = 201
        response.headers = {"X-Data-Source": service_name + " file mock data", "Content-Type": headers['Content-Type']}
    else:
        response.status = 400
        response.data = "Bad Request: no POST body"
    return response


def put_mockdata_url(service_name, conf, url, headers, body, dir_base=dirname(__file__)):
    """
    :param service_name:
        possible "sws", "pws", "book", "hfs", etc.
    """
    # Currently this put method does not return a response body
    response = MockHTTP()
    if body is not None:
        response.status = 204
        response.headers = {"X-Data-Source": service_name + " file mock data", "Content-Type": headers['Content-Type']}
    else:
        response.status = 400
        response.data = "Bad Request: no POST body"
    return response


def delete_mockdata_url(service_name, conf, url, headers, dir_base=dirname(__file__)):
    """
    :param service_name:
        possible "sws", "pws", "book", "hfs", etc.
    """
    # Http response code 204 No Content:
    # The server has fulfilled the request but does not need to return an entity-body
    response = MockHTTP()
    response.status = 204

    return response


def convert_to_platform_safe(dir_file_name):
    """
    :param dir_file_name: a string to be processed
    :return: a string with all the reserved characters replaced
    """
    return re.sub('[?|<>=*,;+&"@]', '_', dir_file_name)



import os
from setuptools import setup, find_packages

BASE_DIR = os.path.abspath(os.path.dirname(__file__))
VERSION_FILE = os.path.join(BASE_DIR, 'resttools', 'VERSION')
with open(VERSION_FILE) as f:
    VERSION = f.readlines()[-1].strip()

with open(os.path.join(BASE_DIR, 'README.md')) as f:
      LONG_DESCRIPTION = f.read()

setup(name='uw-iam-resttools',
      version=VERSION,
      url='https://github.com/UWIT-IAM/iam-resttools',
      author='UW-IT Identity and Access Management',
      author_email='help@uw.edu',
      description='UW-IT REST library for IAM APIs',
      long_description=LONG_DESCRIPTION,
      long_description_content_type='text/markdown',
      license='Apache License, Version 2.0',
      packages=find_packages(),
      include_package_data=True,
      install_requires=['lxml', 'python-dateutil', 'urllib3', 'jinja2', 'six']
      )



import json
import logging
from nose.tools import *

from resttools.gws import GWS
from resttools.exceptions import DataFailureException

import resttools.test.test_settings as settings
import logging.config
logging.config.dictConfig(settings.LOGGING)
logger = logging.getLogger(__name__)


class GWS_Test():

    def __init__(self):
        self.gws = GWS(settings.GWS_CONF)

    def test_search_groups(self):
        groups = self.gws.search_groups(name='2015spr-phys*1', stem='course')
        eq_(len(groups), 199)
        groups = self.gws.search_groups(member='javerage')
        eq_(len(groups), 24)

    def test_get_group(self):
        group = self.gws.get_group_by_id('course_2015spr-phys114a')
        eq_(group.name, 'course_2015spr-phys114a')

    @raises(DataFailureException)
    def test_get_group_members_403(self):
        members = self.gws.get_members('course_2015spr-phys114a')



"""
This is the interface for interacting with the Notify Web Service.
"""
from resttools.dao import NTFYWS_DAO
import json
import logging
logger = logging.getLogger(__name__)


class NTFYWS(object):

    def __init__(self, conf, actas=None):
        self._service_name = conf['SERVICE_NAME']
        self.dao = NTFYWS_DAO(conf)

    def send_message(self, eppn, number, message, type='text'):
        """
        Sends a text (or voice) message to the phone number
        """
        if type == 'text':
            mtype = 'uw_direct_phone_sms'
            ts = 'Text'
        elif type == 'voice':
            mtype = 'uw_direct_phone_voice'
            ts = 'Say'
        else:
            return 400

        data = {'Dispatch': {
            'MessageType': mtype, 'Content': {
                'Recipient': eppn,
                'PhoneNumber': number,
                ts: message
                }
        }}

        url = "/%s/v1/dispatch" % (self._service_name)
        resp = self.dao.postURL(url, {"Content-Type": "application/json"}, json.dumps(data))

        return resp.status



"""
Contains IRWS DAO implementations.
"""
from resttools.mock.mock_http import MockHTTP
import json
from resttools.dao_implementation.live import get_con_pool, get_live_url
from resttools.dao_implementation.mock import get_mockdata_url

import logging
logger = logging.getLogger(__name__)


class File(object):
    """
    The File DAO implementation returns generally static content.  Use this
    DAO with this configuration:

    """
    _max_pool_size = 5
    _cache_db = {}

    def __init__(self, conf):
        self._conf = conf
        if 'MAX_POOL_SIZE' in conf:
            self._max_pool_size = conf['MAX_POOL_SIZE']

    def getURL(self, url, headers):
        logger.debug('file irws get url: ' + url)
        if url in File._cache_db:
            logger.debug('using cache')
            response = MockHTTP()
            response.data = File._cache_db[url]
            response.status = 200
            return response
        response = get_mockdata_url("irws", self._conf, url, headers)
        if response.status == 404:
            logger.debug('status 404')
            response.data = '{"error": {"code": "7000","message": "No record matched"}}'
        return response

    def putURL(self, url, headers, body):
        logger.debug('file irws put url: ' + url)

        response = get_mockdata_url("irws", self._conf, url, headers)
        if response.status != 404 or url in File._cache_db:
            # try set in cache
            cache_data = json.loads(File._cache_db.get(url, None) or
                                    response.data)
            put_data = json.loads(body)
            key, items = next(iter(put_data.items()))
            put_section = items[0]
            if key == 'name':
                # fake a cname update
                name_parts = [put_section.get('preferred_{}name'.format(x), '')
                              for x in ('f', 'm', 's')]
                put_section['preferred_cname'] = ' '.join(x for x in name_parts if x)
            # update the put data and leave everything else in place
            cache_data[key][0].update(put_section)
            File._cache_db[url] = json.dumps(cache_data)
            # return an irws-style put response
            response.data = '{"cached": {"code": "0000","message": "put cached in mock data"}}'
            response.status = 200

        return response

    def postURL(self, url, headers, body):
        return self.putURL(url, headers, body)

    def deleteURL(self, url, headers):
        logger.debug('file irws delete url: ' + url)
        response = get_mockdata_url("irws", self._conf, url, headers)
        # no status for deletes from mock
        response.status = 200
        return response


class Live(object):
    """
    This DAO provides real data.  It requires further configuration, (conf)
    """
    _max_pool_size = 5

    def __init__(self, conf):
        self._conf = conf
        if 'MAX_POOL_SIZE' in conf:
            self._max_pool_size = conf['MAX_POOL_SIZE']

    # class attribute means you can't switch hosts/keys
    pool = None

    def getURL(self, url, headers):
        return get_live_url(self._get_pool(), 'GET',
                            self._conf['HOST'],
                            url, headers=headers,
                            service_name='irws')

    def deleteURL(self, url, headers):
        return get_live_url(self._get_pool(), 'DELETE',
                            self._conf['HOST'],
                            url, headers=headers,
                            service_name='irws')

    def putURL(self, url, headers, body):
        return get_live_url(self._get_pool(), 'PUT',
                            self._conf['HOST'],
                            url, headers=headers, body=body,
                            service_name='irws')

    def postURL(self, url, headers, body):
        return get_live_url(self._get_pool(), 'POST',
                            self._conf['HOST'],
                            url, headers=headers, body=body,
                            service_name='irws')

    def _get_pool(self):
        if not Live.pool:
            Live.pool = get_con_pool(self._conf['HOST'],
                                     self._conf['KEY_FILE'],
                                     self._conf['CERT_FILE'],
                                     self._conf['CA_FILE'],
                                     max_pool_size=self._max_pool_size,
                                     verify_https=self._conf.get('VERIFY_HOST', True))
        return Live.pool



"""
This is the interface for interacting with the UW NetID Web Service.
"""
from resttools.dao import NWS_DAO
from resttools.models.nws import UWNetIdAdmin, UWNetIdPwInfo
from resttools.exceptions import DataFailureException
from six.moves.urllib.parse import quote_plus
import json
import logging
logger = logging.getLogger(__name__)

# note.  Some messages from nws set pw
# 409, "That's not a good password because you used that password in the not too distant past."
# 409, "That's not a good password because it is your username."
# 409, "That's not a good password because it does not contain enough different character classes
#       (uppercase, lowercase, digits, punctuation)."
# 409, "That's not a good password because it needs more complex capitalization."
#


class NWS(object):

    def __init__(self, conf, actas=None):
        service_name = conf['SERVICE_NAME']
        version = conf.get('VERSION', 'v1')
        self._base_url = '/{}/{}'.format(service_name, version)
        self._pw_action = 'Set'
        if 'PASSWORD_ACTION' in conf:
            self._pw_action = conf['PASSWORD_ACTION']
        self.dao = NWS_DAO(conf)

    def get_netid_admins(self, netid):
        """
        Returns a list of NetidAdmin objects for the netid
        """
        url = "%s/uwnetid/%s/admin" % (self._base_url, quote_plus(netid))
        response = self.dao.getURL(url, self._headers({"Accept": "application/json"}))

        if response.status != 200:
            raise DataFailureException(url, response.status, response.data)

        return self._admins_from_json(response.data)

    def get_netid_pwinfo(self, netid):
        """
        Returns NetidPwINfo object for the netid
        """
        url = "%s/uwnetid/%s/password" % (self._base_url, quote_plus(netid))
        response = self.dao.getURL(url, self._headers({"Accept": "application/json"}))

        if response.status != 200:
            raise DataFailureException(url, response.status, response.data)

        return self._pwinfo_from_json(response.data)

    def set_netid_pw(self, netid, password, auth, action=None):
        """
        Sets password for netid
        """

        if action is None:
            action = self._pw_action

        url = "%s/uwnetid/%s/password" % (self._base_url, quote_plus(netid))
        data = {'action': action, 'newPassword': password, 'uwNetID': netid, 'authMethod': auth}
        response = self.dao.postURL(url, {"Content-type": "application/json"}, json.dumps(data))

        if response.status >= 500:
            raise DataFailureException(url, response.status, response.data)
        obj = json.loads(response.data)
        if 'result' not in obj:
            # probably a mock data issue
            return (500, "Unable to set password.")
        return (obj['result'], obj['message'])

    def _admins_from_json(self, data):
        adminobj = json.loads(data)
        admins = [UWNetIdAdmin(name=admin.get('name', ''),
                               role=admin.get('role', ''),
                               type=admin.get('type', ''))
                  for admin in adminobj.get('adminList', [])]
        return admins

    def _pwinfo_from_json(self, data):
        infoobj = json.loads(data)
        if 'minimumLength' in infoobj:
            return UWNetIdPwInfo(min_len=infoobj.get('minimumLength', None),
                                 last_change=infoobj.get('lastChange', None),
                                 kerb_status=infoobj.get('kerbStatus', None))
        else:
            return None

    def _headers(self, headers):
        # could auto-add headers here
        return headers

    def _add_header(self, headers, header, value):
        if not headers:
            return {header: value}

        headers[header] = value
        return headers



# various nose tests

from nose.tools import *

import resttools.test.test_settings as settings

from resttools.test.irws import IRWS_Test
from resttools.test.nws import NWS_Test
from resttools.test.gws import GWS_Test
